#![warn(clippy::perf)]
#![warn(clippy::all)]

pub mod is_prime;
pub mod primality_test;

use crate::is_prime::prime_utils;
use num::{BigUint, One};
use once_cell::sync::Lazy;
use std::{
    io::Write,
    sync::{
        atomic::{AtomicU64, AtomicUsize},
        Arc, RwLock,
    },
    thread,
    time::SystemTime,
};

#[cfg(target_os = "windows")]
use thread_priority::{set_current_thread_priority, ThreadPriority};

fn p(n: f64) -> f64 {
    let ln_n = n.ln();
    let ln_ln_n = ln_n.ln();
    let ln_ln_ln_n = ln_ln_n.ln();

    n * (ln_n + ln_ln_n - 1.0 + (ln_ln_ln_n - 2.0) / ln_n
        - ((ln_ln_n).powi(2) - 6.0 * ln_ln_n + 11.0) / (2.0 * 2.0f64.log2() * ln_n)
        + ((ln_ln_n / ln_n).powi(3)) * (1.0 / ln_n))
}

fn to_usize(f: f64) -> usize {
    let rounded_num: i64 = f.round() as i64;
    if rounded_num >= 0 {
        rounded_num as usize
    } else {
        0
    }
}

fn to_f64(u: usize) -> f64 {
    f64::from_bits(u as u64)
}

#[cfg(target_os = "windows")]
fn try_set_thread_priority() {
    if let Err(e) = set_current_thread_priority(ThreadPriority::Max) {
        eprintln!("[WARN] failed to set thread priority to max");
        eprintln!("{e}");
    }
}

fn write_primes(primes: &[BigUint]) {
    let mut file = std::fs::File::create("primes.json").unwrap();
    file.write_all(serde_json::to_string(primes).unwrap().as_bytes())
        .unwrap();
}

const N: i64 = 1_000_000;
const NF: f64 = N as f64;
const NU: usize = N as usize;
const THREADS: usize = 30;
const SPLITTER: u64 = N as u64 / THREADS as u64;

fn worker(
    workloads: usize,
    i: usize,
    progress: &AtomicU64,
    arcclone: &Arc<RwLock<Vec<BigUint>>>,
    threads_done: &AtomicUsize,
) {
    #[cfg(target_os = "windows")]
    try_set_thread_priority();

    static INC: once_cell::sync::Lazy<BigUint> = Lazy::new(BigUint::one);
    static ONE_THOUSAND: once_cell::sync::Lazy<BigUint> =
        Lazy::new(|| BigUint::from(SPLITTER * 10));
    static ONE_HUNDRED: once_cell::sync::Lazy<BigUint> = Lazy::new(|| BigUint::from(SPLITTER));
    static TEN: once_cell::sync::Lazy<BigUint> = Lazy::new(|| BigUint::from(10u8));

    let work_uint = BigUint::from(workloads);
    let mut number = BigUint::from(i) * &work_uint;

    let max = BigUint::from(i + 1) * work_uint;
    let mut primecache = Vec::with_capacity(NU);
    let mut primecache_outside_len = 0;
    let mut global_primes = arcclone.read().unwrap();
    loop {
        number += &*INC;

        if prime_utils::is_prime(&number, &global_primes) {
            primecache.push(number.clone());
            progress.fetch_add(1, std::sync::atomic::Ordering::Relaxed);
        }

        if (&number * &*ONE_THOUSAND) / &max / &*TEN == &number / &max * &*ONE_HUNDRED {
            drop(global_primes);

            let mut new_primecache = arcclone.write().unwrap();

            new_primecache.extend_from_slice(&primecache[primecache_outside_len..]);
            new_primecache.sort_unstable();

            primecache.clone_from(&new_primecache);
            primecache_outside_len = new_primecache.len();

            drop(new_primecache);

            global_primes = arcclone.read().unwrap();
        }
        if number == max {
            break;
        }
    }

    drop(global_primes);

    let mut new_primecache = arcclone.write().unwrap();

    new_primecache.extend_from_slice(&primecache[primecache_outside_len..]);
    new_primecache.sort_unstable();

    threads_done.fetch_add(1, std::sync::atomic::Ordering::Relaxed);
}

fn main() {
    let sys_time = SystemTime::now();

    let nthprime_approx: usize = to_usize(p(NF * 1.02));

    let workloads = nthprime_approx / THREADS;

    let primecache: Vec<BigUint> = Vec::with_capacity(NU);
    let primemutex = std::sync::RwLock::new(primecache);
    let primearc = std::sync::Arc::new(primemutex);

    let progress_val = AtomicU64::new(0);
    let progress = &progress_val;

    let threads_done_val = AtomicUsize::new(0);
    let threads_done = &threads_done_val;

    thread::scope(|scope| {
        for i in 0..THREADS {
            let arcclone = primearc.clone();
            scope.spawn(move || worker(workloads, i, progress, &arcclone, threads_done));
        }

        loop {
            let n = progress.load(std::sync::atomic::Ordering::Relaxed);
            let threads_completed = threads_done.load(std::sync::atomic::Ordering::Relaxed);

            println!(
                "Progress: {}/{} ({:.0}%) | {}",
                n,
                N,
                n as f64 / NF * 100.0,
                threads_completed
            );
            std::thread::sleep(std::time::Duration::from_millis(500));

            if threads_completed == THREADS {
                break;
            }
        }
    });

    let mut primes = primearc.write().unwrap();

    #[allow(clippy::cast_possible_wrap)]
    let removed: i32 = if primes.len() > NU {
        let rem = primes.len() - NU;
        primes.truncate(NU);
        rem as i32
    } else {
        0
    };

    let added = NU - primes.len();
    if primes.len() < NU {
        static INC: once_cell::sync::Lazy<BigUint> = Lazy::new(BigUint::one);

        println!("[INFO] less primes than expected");
        let mut number = BigUint::from(THREADS) * BigUint::from(workloads);
        loop {
            number += &*INC;
            if prime_utils::is_prime(&number, &primes) {
                primes.push(number.clone());
            }
            if primes.len() == NU {
                break;
            }
        }
    }

    let difference = sys_time.elapsed().unwrap();

    assert_eq!(primes.len(), NU);

    println!(
        "primes added: {added} | % added: {}% | removed: {removed}",
        to_f64(added) / to_f64(primes.len()) * 100.0
    );
    println!("time: {difference:?}");
    println!("writing primes...");

    write_primes(&primes);
}
